//----------------------------------------------------------------------------
//  Body: CDIMServer.cpp, v 1.0                               2007/07/18    //
//                                                                          //
//  Goofie TPC ALICE Program Library                                        //
//  Copyright (C) ALICE collaboration, 2007                                 //
//                                                                          //
//  Last modified in CVS                                                    //
//  Author: D. Antonczyk                                                    //
//----------------------------------------------------------------------------

#include <iostream>
#include <string>

#include "CMacroDef.h"
#include "SConfigData.h"
#include "CGasAnaResults.h"
#include "CDIMServer.h"

// error coding rules
// 9000 for dim XXXX error code i.e. -90001001, - for errors, + for warnings

using namespace std;

// action and states from _tpcStateDef
const int CDIMServer::kActionStop    = 3;   // not used
const int CDIMServer::kActionStart   = 15;  // not used

const int CDIMServer::kActionGoOn    = 9;
const int CDIMServer::kActionGoOff   = 1;
const int CDIMServer::kActionRecover = 13;
const int CDIMServer::kStateOff      = 4;
const int CDIMServer::kStateOn       = 11;
const int CDIMServer::kStateError    = 13;


CDIMServer::CDIMServer() : DimServer(),
			   fkState(kStateOn),
			   fkAction(kActionGoOn)
{
   // dim service
   SetToNullDataPoint();
   
   // dim service FSM
   fskState                  = NULL;
   // dim action FSM
   fckAction                 = NULL;
   // buffer
   fGasAnalysisResults       = NULL; 
}

void CDIMServer::SetToNullDataPoint()
{
   fsFieldCageHV             = NULL;
   fsPickupHV                = NULL;
   fsPeakPosNear             = NULL;
   fsPeakPosFar              = NULL;
   fsPeakAreaNear            = NULL;
   fsPeakAreaFar             = NULL;
   fsTemp1                   = NULL;
   fsTemp2                   = NULL;
   fsPressure                = NULL;
   fsGainNear                = NULL;
   fsGainNearCorrected       = NULL;
   fsGainFar                 = NULL;
   fsGainFarCorrected        = NULL;
   fsFractionCO2             = NULL;
   fsFractionN2              = NULL;
   fsDriftVelocity           = NULL;
   fsDriftVelocityCorrected  = NULL;
   fsCorrectionFactor        = NULL;
   fsNumberOfDischarge       = NULL;
}

CDIMServer::~CDIMServer()
{
   StopServer();
}

void CDIMServer::SetDataPointer(const CGasAnaResults *_res)
{
   fGasAnalysisResults = _res;
}
    
int CDIMServer::CreateServer(string _serverName)
{
   int err = 0;
   // check whether the servername is not already set properly   
   if(_serverName.compare(fserverName) != 0) {
     fserverName = _serverName+".GOOFIE";     // set server name;
   }
   try {
      // not alow to add any new service after server started
      autoStartOff();
      // create name space
      if(CreateServerNameSpace()) throw (int)-90000000;
      // create server
      start(fserverName.c_str());
   } catch (int _err) {
      cerr << "Error in <CDIMServer::CreateServer>: creation of dim server unsuccessful, code="<< _err << flush << endl;
      err = _err;
   }
   // set action to start data update
   fkAction = kActionGoOn;
   // set state On
   SetState(kStateOn);
   UpdateStates();

   DEBUGINFO("Debuge Info => CDIMServer::CreateServer => DIM server created",err)
   return err;
}

int CDIMServer::RunServer()
{
   int err = 0;
   if(GetAction()==kActionGoOff) {          // stop data aquisation
      if(GetState() != kStateOff) {
	 SetState(kStateOff);                  // off
	 UpdateStates();
      }
      DEBUGINFO("Debuge Info => CDIMServer::RunServer() => stop data updating = ", err);
      
   } else if(GetAction()==kActionGoOn) {    // start data aquisation

      if(GetState() != kStateOn) {
	 SetState(kStateOn);               // state on
	 UpdateStates();
      }

      DEBUGINFO("Debuge Info => CDIMServer::RunServer() => data updating = ", err);

      err = UpdateValues();             // update data points
      if(err) {
	 DEBUGINFO("Debuge Info => CDIMServer::RunServer() => error value = ", err);
	 if(GetState() != kStateError) {
	    SetState(kStateError);         // state error
	    UpdateStates();
	 }
      }
   } else if((GetAction()==kActionRecover) && (GetState() == kStateError)) {
      Recovery();
      DEBUGINFO("Debuge Info => CDIMServer::RunServer() => Recovery from the error ",0);
   }
   // test
   //cout << "DNS node = " << getDnsNode() << " port = " << getDnsPort() << endl;
   //cout << "error is = " << err << endl;
   
   return err;
}

int CDIMServer::Recovery()
{
   // stop and delete all data points
   StopServer();
   // create and start DIM server
   CreateServer(fserverName);
   
   return 0;
}

int CDIMServer::StopServer()
{
   int err     = 0;
   int _err    = 0;
   int counter = -1;
   
   try {
      // stop server 
      stop();

      // delete data Points
      while (!fListOfDataPoints.empty()) {
	 ++counter;
	 _err = (-90000300) - counter;
	 
	 DimService *tempDimService = fListOfDataPoints.back();

	 if(tempDimService) {
	    delete tempDimService;
	    fListOfDataPoints.pop_back();
	 } else throw _err;
      }

      // clear data pointers
      SetToNullDataPoint();
      
      // delete dim service
      if(fskState) {
	 delete fskState;
	 fskState = NULL;
      } else throw (int)-90000400;

      // delete dim commands
      if(fckAction) {
	 delete fckAction;
	 fckAction = NULL;
      } else throw (int)-90000401;
      
   } catch(int __err) {
      cerr << "Error in <CDIMServer::StopServer>: error code=" << _err << flush << endl;
      err = __err;
   }
   return err;
}

int CDIMServer::CreateServerNameSpace()
{
   int err = 0;
   string tempServiceName;
  
   try {
      // data Points
      if(!fsFieldCageHV) {
	 tempServiceName = SConfigData::fDIMServerPrefix + ".GAS.GOOFIE.FCageHV";
	 fsFieldCageHV = new DimService(tempServiceName.c_str(), fDimDataPoints.fFieldCageHV);
	 fListOfDataPoints.push_back(fsFieldCageHV);
      } else throw (int)-90000400;

      if(!fsPickupHV) {
	 tempServiceName = SConfigData::fDIMServerPrefix + ".GAS.GOOFIE.PickupHV";
	 fsPickupHV = new DimService(tempServiceName.c_str(), fDimDataPoints.fPickupHV);
	 fListOfDataPoints.push_back(fsPickupHV);
      } else throw (int)-90000401;

      if(!fsPeakPosNear) {
	 tempServiceName = SConfigData::fDIMServerPrefix + ".GAS.GOOFIE.PeakPosNear";
	 fsPeakPosNear = new DimService(tempServiceName.c_str(), fDimDataPoints.fPeakPosNear);
	 fListOfDataPoints.push_back(fsPeakPosNear);
      } else throw (int)-90000402;

      if(!fsPeakPosFar) {
	 tempServiceName = SConfigData::fDIMServerPrefix + ".GAS.GOOFIE.PeakPosFar";
	 fsPeakPosFar = new DimService(tempServiceName.c_str(), fDimDataPoints.fPeakPosFar);
	 fListOfDataPoints.push_back(fsPeakPosFar);
      } else throw (int)-90000403;
      
      if(!fsPeakAreaNear) {
	  tempServiceName = SConfigData::fDIMServerPrefix + ".GAS.GOOFIE.PeakAreaNear";
	  fsPeakAreaNear = new DimService(tempServiceName.c_str(), fDimDataPoints.fPeakAreaNear);
	  fListOfDataPoints.push_back(fsPeakAreaNear);
      } else throw (int)-90000404;

      if(!fsPeakAreaFar) {
	 tempServiceName = SConfigData::fDIMServerPrefix + ".GAS.GOOFIE.PeakAreaFar";
	 fsPeakAreaFar =  new DimService(tempServiceName.c_str(), fDimDataPoints.fPeakAreaFar);
	 fListOfDataPoints.push_back(fsPeakAreaFar);
      } else throw (int)-90000405;

      if(!fsTemp1) {
	 tempServiceName = SConfigData::fDIMServerPrefix + ".GAS.GOOFIE.TemperatureS1";
	 fsTemp1 =  new DimService(tempServiceName.c_str(), fDimDataPoints.fTemp1);
	 fListOfDataPoints.push_back(fsTemp1);
      } else throw (int)-90000406;

      if(!fsTemp2) {
	 tempServiceName = SConfigData::fDIMServerPrefix + ".GAS.GOOFIE.TemperatureS2";
	 fsTemp2 =  new DimService(tempServiceName.c_str(), fDimDataPoints.fTemp2);
	 fListOfDataPoints.push_back(fsTemp2);
      } else throw (int)-90000407;

      if(!fsPressure) {
	 tempServiceName = SConfigData::fDIMServerPrefix + ".GAS.GOOFIE.Pressure";
	 fsPressure = new DimService(tempServiceName.c_str(), fDimDataPoints.fPressure);
	 fListOfDataPoints.push_back(fsPressure);
      } else throw (int)-90000408;
      
      if(!fsGainNear) {
	 tempServiceName = SConfigData::fDIMServerPrefix + ".GAS.GOOFIE.GainNear";
	 fsGainNear = new DimService(tempServiceName.c_str(), fDimDataPoints.fGainNear);
	 fListOfDataPoints.push_back(fsGainNear);
      } else throw (int)-90000409;

      if(!fsGainNearCorrected) {
	 tempServiceName = SConfigData::fDIMServerPrefix + ".GAS.GOOFIE.GainNearCorrected";
	 fsGainNearCorrected = new DimService(tempServiceName.c_str(), fDimDataPoints.fGainNearCorrected);
	 fListOfDataPoints.push_back(fsGainNearCorrected);
      } else throw (int)-90000410;
      
      if(!fsGainFar) {
	 tempServiceName = SConfigData::fDIMServerPrefix + ".GAS.GOOFIE.GainFar";
	 fsGainFar = new DimService(tempServiceName.c_str(), fDimDataPoints.fGainFar);
	 fListOfDataPoints.push_back(fsGainFar);
      } else throw (int)-90000411;

     if(!fsGainFarCorrected) {
	 tempServiceName = SConfigData::fDIMServerPrefix + ".GAS.GOOFIE.GainFarCorrected";
	 fsGainFarCorrected = new DimService(tempServiceName.c_str(), fDimDataPoints.fGainFarCorrected);
	 fListOfDataPoints.push_back(fsGainFarCorrected);
      } else throw (int)-90000412;
      
      if(!fsFractionCO2) {
	 tempServiceName = SConfigData::fDIMServerPrefix + ".GAS.GOOFIE.CO2";
	 fsFractionCO2 = new DimService(tempServiceName.c_str(), fDimDataPoints.fFractionCO2);
	 fListOfDataPoints.push_back(fsFractionCO2);
      } else throw (int)-90000413;
      
      if(!fsFractionN2) {
	 tempServiceName = SConfigData::fDIMServerPrefix + ".GAS.GOOFIE.N2";
	 fsFractionN2 = new DimService(tempServiceName.c_str(), fDimDataPoints.fFractionN2);
	 fListOfDataPoints.push_back(fsFractionN2);
      } else throw (int)-90000414;

      if(!fsDriftVelocity) {
	 tempServiceName = SConfigData::fDIMServerPrefix + ".GAS.GOOFIE.DriftVelocity";
	 fsDriftVelocity = new DimService(tempServiceName.c_str(), fDimDataPoints.fDriftVelocity);
	 fListOfDataPoints.push_back(fsDriftVelocity);
      } else throw  (int)-90000415;

       if(!fsDriftVelocityCorrected) {
	 tempServiceName = SConfigData::fDIMServerPrefix + ".GAS.GOOFIE.DriftVelocityCorrected";
	 fsDriftVelocityCorrected = new DimService(tempServiceName.c_str(), fDimDataPoints.fDriftVelocityCorrected);
	 fListOfDataPoints.push_back(fsDriftVelocityCorrected);
      } else throw  (int)-90000416;

       if(!fsCorrectionFactor) {
	 tempServiceName = SConfigData::fDIMServerPrefix + ".GAS.GOOFIE.CorrectionFactor";
	 fsCorrectionFactor = new DimService(tempServiceName.c_str(), fDimDataPoints.fCorrectionFactor);
	 fListOfDataPoints.push_back(fsCorrectionFactor);
      } else throw  (int)-90000417;
      
      if(!fsNumberOfDischarge) {
	 tempServiceName = SConfigData::fDIMServerPrefix + ".GAS.GOOFIE.NumberOfSparks";
	 fsNumberOfDischarge = new DimService(tempServiceName.c_str(), fDimDataPoints.fNumberOfDischarge);
	 fListOfDataPoints.push_back(fsNumberOfDischarge);
      } else throw (int)-90000418;
      //
      // FSM
      // dim service FSM
      if(!fskState) {
	 tempServiceName = SConfigData::fDIMServerPrefix + ".GAS.GOOFIE.kState";
	 fskState = new DimService(tempServiceName.c_str(), fkState);
      } else throw (int)-90000419;
      //
      // command == actions
      if(!fckAction) {
	 tempServiceName = SConfigData::fDIMServerPrefix + ".GAS.GOOFIE.kAction";
	 fckAction = new DimCommand(tempServiceName.c_str(), "I", this);
      } else throw (int)-90000420;
      
   } catch (int _err) {
      cerr << "Error in <CDIMServer::CreateServerNameSpace>: initialization of server name space, code=" << _err << flush << endl;
      err = _err;
   }
   return err;
}

int CDIMServer::UpdateValues()
{
   int err     = 0;
   int _err    = 0;
   int counter = -1;
   try {
      // set new values
      if(fGasAnalysisResults) {
	 fDimDataPoints = *fGasAnalysisResults;
      } else throw (int)-90000500;
      
      // update service
      for(fitForListOfDataPoint=fListOfDataPoints.begin(); fitForListOfDataPoint!=fListOfDataPoints.end(); ++fitForListOfDataPoint) {
	 DimService *tempDimService = *fitForListOfDataPoint;	 
	 if(tempDimService) {
	    tempDimService->updateService();
	    ++counter;
	    _err = (-90000500) - counter;
	 } else throw _err;
      }
      
   } catch (int __err) {
      cerr << "Error in <CDIMServer::UpdateValues>: pointer equal to NULL, code=" << __err << flush << endl;
      err = __err;
   }
   return err;
}

int CDIMServer::UpdateStates()   
{
   // state list
   //------------------------------------
   // name                         value
   //------------------------------------
   // kStateON                      11
   // kStateOff                     4
   // kStateError                   13
   
   int err = 0;
   try {
      fskState->updateService();
   } catch (...) {
      cerr << "Error in <CDIMServer::UpdateStates>: unknown error "<< flush << endl;
      err = -1;
   }
   return err; 
}

void CDIMServer::commandHandler()
{
   // action list
   //------------------------------------
   // name                         value
   //------------------------------------
   // kActionStart                  15       // not used 
   // kActionStop                   3        // not used
   // kActionGoOn                   9
   // kActionGoOff                  1
   // kActionRecover                13
  
   DimCommand *tcomm = getCommand();
   
   if(tcomm==fckAction) {
      fkAction = fckAction->getInt();
      DEBUGINFO("Debuge Info => CDIMServer::commandHandler() => recived action no. = ", fkAction);
   }
}
