//----------------------------------------------------------------------------
//  Body: COnLineAna.cpp, v 1.0                               2007/04/16    //
//                                                                          //
//  Goofie TPC ALICE Program Library                                        //
//  Copyright (C) ALICE collaboration, 2007                                 //
//                                                                          //
//  Last modified in CVS                                                    //
//  Author: D. Antonczyk                                                    //
//----------------------------------------------------------------------------


#include <iostream>
#include <sstream>
#include <pthread.h>

#include "CMacroDef.h"
#include "CService.h"
#include "COnLineExeption.h"
#include "CDataStructure.h"
#include "SConfigData.h"
#include "CDIMServer.h"
#include "CGasAnaResults.h"
#include "CRootFileHandler.h"
#include "AliGoofieGasAnalysis.h"
#include "AliGoofieEventInfo.h"

#include "COnLineAna.h"

// TODO:
//

using namespace std;


// ------------- START - static functions - START ---------------------------------
// thread function - main analysis loop
// all necessary functionality should be added in this function 
void *COnLineAna::MakeAnalysis(void* arg)
{
   COnLineAna *fONA = reinterpret_cast<COnLineAna*>(arg);
   
   // set cancelation state
   pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
   // wait while creation of thread is over.
   usleep(50000);
   
   try { 
      
      while(fONA->fWrapper->GetRunONAStatus()) {
	 // cancelation point
	 pthread_testcancel();

	 // receiving pack from queue
	 fONA->GetAndAnalysePack();
	 
	 // Update DIM server
	 fONA->MoveResultsToCSDimServer();

	 // Fill/Copy ROOT file to storage place
	 fONA->FillAndCopyRootFile();
      }
   }
   catch(COnLineExeption *eVal) {
      // for debuging only
      DEBUGINFO("Debuge Info => COnLineAna::MakeAnalysis => Exeption !!!! ",0)
      fONA->fExeptionStatus->PrintExeptionError();
   }
   
   return NULL;
}


//---------constructor----------
COnLineAna::COnLineAna()
{
  COnLineAna(NULL);
  cout << "Warning from <COnLineAna::COnLineAna()>: CService pointer is set to NULL" << flush << endl;
}


COnLineAna::COnLineAna(CService *service)
{
   // set pointer for CService obj.
   fWrapper = service;

   // error / exeption handling
   fExeptionStatus  = new COnLineExeption();

   // objects for gas analysis
   fGasAnalysis  = new AliGoofieGasAnalysis();
  
   // local results buffor
   fGasAnalysisResults = new CGasAnaResults();

   // object for handling with DIM client/server app.
   fDimServerHandler = new CDIMServer();  
   // set data address for DIM server
   fDimServerHandler->SetDataPointer(fGasAnalysisResults);

   // storing place for events
   fPack    = new deque<CDataStructure>(SConfigData::fEventsInPack);

   // object for handling with root files and acquisition
   fRootFileHandler = new CRootFileHandler();
   
   // local copy just to avoid access to the global variable
   // by two or more threads at the same time
   fLocalCopyOfEventsInPack = SConfigData::fEventsInPack;

   fCurrentPackNumber = 0;
	 
   DEBUGINFO("Debuge Info => COnLineAna object created",0)  
}


//---------destructor----------
COnLineAna::~COnLineAna()
{
   // close ONA thread
   pthread_exit(NULL);
   
   if(fPack) {
      fPack->clear();
      delete fPack;
      fPack=NULL;
   }
   if(fRootFileHandler) {
      delete fRootFileHandler;
      fRootFileHandler = NULL;
   }
   if(fExeptionStatus) {
      delete fExeptionStatus;
      fExeptionStatus=NULL;
   }
   if(fGasAnalysis) {
     delete fGasAnalysis;
     fGasAnalysis = NULL;
   }
   DEBUGINFO("Debuge Info => COnLineAna object destroyed",0)
}


void COnLineAna::InitializeSoftware(const string _configFile) throw(COnLineExeption)
{
   unsigned int err = 0;
  
   try {
      // set and configure analysis part
      fGasAnalysis->SetConfiguration(_configFile);

      // create server (open conetcion with DNS) 
      if(fDimServerHandler->CreateServer(SConfigData::fDIMServerPrefix)) {
	 err = 0x1000000;
	 throw err;
      }
            
   } catch (unsigned int _err) {
      fExeptionStatus->SetExeptionError(_err, "<COnLineAna::InitializeSoftware()>: Initialization of the analysis part failed");
      throw fExeptionStatus;
   }
}


pthread_t *COnLineAna::StartONA(const string _configFile)
{
 unsigned int err = 0;
   try {
      // initialization
      InitializeSoftware(_configFile);      // throw COnLineExeption
            
      pthread_attr_t  attr_ona;
      // setting thread attribute
      err = pthread_attr_init(&attr_ona);
      DEBUGINFO("Debuge Info => COnLineAna::StartONA => thread attribute initialization",err)
      if(err) throw err;

      // attribute ...
      pthread_attr_setdetachstate(&attr_ona, PTHREAD_CREATE_JOINABLE);
      // stack size
      size_t stacksize;
      err = pthread_attr_getstacksize (&attr_ona, &stacksize);
      DEBUGINFO("Debuge Info => COnLineAna::StartONA => getting stack size",err)
      if(err) throw err;
      stacksize = (size_t)4.0*stacksize;
      err = pthread_attr_setstacksize (&attr_ona, stacksize);
      DEBUGINFO("Debuge Info => COnLineAna::StartONA => setting new stack size",err)
      if(err) throw err;
      
      // create thread
      err = pthread_create(&tid_ona, &attr_ona, MakeAnalysis, (void*)this);
      DEBUGINFO("Debuge Info => COnLineAna::StartONA => thread created",err)
      if(err) throw err;
      // delete thread attribute obj
      err = pthread_attr_destroy(&attr_ona);
      DEBUGINFO("Debuge Info => COnLineAna::StartONA => thread attribute destroyed",err)
      if(err) throw err;
      
   } catch (unsigned int _err) {
      fExeptionStatus->SetExeptionError(_err, "<COnLineAna::StartONA()>: Initialization of the thread failed");
   }
   return &tid_ona;
}


void COnLineAna::GetAndAnalysePack() throw(COnLineExeption)
{
   unsigned int err = fWrapper->GetPackFromQueue(fPack);
   if(err) {
      fExeptionStatus->SetExeptionError(err, "<COnLineAna::GetAndAnalysePack()>: Getting pack from queue failed");
      throw fExeptionStatus;
   }
   DEBUGINFO("Debuge Info => COnLineAna::GetAndAnalysePack => pack recived from queue",err)

   cout << "START -> Current Pack size = " << fAnalysisPack.size() << flush << endl;

   // ------------------ Gas analysis --------------------------------
   // prepare list of event for analysis 
   for(unsigned int pos = 0; pos < fPack->size(); ++pos) {
      fAnalysisPack.push_back(fPack->at(pos));
   }
   cout << "Debuge Info => COnLineAna::GetAndAnalysePack =>  pack added to the analysis pack" << endl;
   //DEBUGINFO("Debuge Info => COnLineAna::GetAndAnalysePack =>  pack added to the analysis pack",err)
   
   if(fAnalysisPack.size() >= ((unsigned int)SConfigData::fEventsUsedForAnalysis)) {
     
      cout << "------------------------------ Starting Analysis ------------------ " << endl;
   
      // start analysis
      fGasAnalysis->AnalyzePack(fAnalysisPack);

      // clear internal event list
      fGasAnalysis->ClearEventList();
  
      // display gas analysis results
      fGasAnalysis->PrintResults(SConfigData::fPrintGasAnalysisResults);

      //cout << "results " << DumpVectorToString() << endl;

      fGasAnalysisResults->SetResults(fGasAnalysis->GetGasAnaResults());
    
      cout << "Going to remove last pack from list of event for analysis, size = " << fAnalysisPack.size() << endl;
      
      // remove first pack from the deque
      for(int pos=0; pos < SConfigData::fEventsInPack; ++pos) {
         fAnalysisPack.pop_front();
      }
                        
      DEBUGINFO("Debuge Info => COnLineAna::GetAndAnalysePack =>  pack analyzed",err)
   }
}


void COnLineAna::MoveResultsToCSDimServer() throw(COnLineExeption)
{
   unsigned int err = 0;
   // add checking for states of application
    
   // update server
   err = fDimServerHandler->RunServer();
   if(err) {
      fExeptionStatus->SetExeptionError(err, "<COnLineAna::MoveResultsToCSDimServer()>: Update of DIM server failed");
      throw fExeptionStatus;
   }
   DEBUGINFO("Debuge Info => COnLineAna::MoveResultsToCSDimServer() => DIM server Updated", err)
}


void COnLineAna::FillAndCopyRootFile() throw(COnLineExeption)
{
   if(SConfigData::fStoreRawData) {
      // create tree and file 
      if(fCurrentPackNumber==0) {
	 fRootFileHandler->CreateOutputFile();
	 fRootFileHandler->Initialize();
      }
      // fill tree
      fRootFileHandler->AddPackToTree(fGasAnalysisResults, *fPack);
      // next pack has been stored
      ++fCurrentPackNumber;
      // if true write tree and close file
      if(fCurrentPackNumber==SConfigData::fRawDataSize) {
	 // save root file to location given in SConfigData object
	 fRootFileHandler->SaveDataLocally();
	 fCurrentPackNumber = 0;
      }
   }
   DEBUGINFO("Debuge Info =>  COnLineAna::FillAndCopyRootFile() => Pack saved in tree", 0)
}


// ------------------------- Test Functions ---------------------------------- //
string COnLineAna::DumpVectorToString()
{
  stringstream  sRetName;
  sRetName.clear();
  
  if(fPack) {
    // lock
    unsigned int err = pthread_mutex_lock(&fWrapper->fmutex);
    DEBUGINFO("Debuge Info => COnLineAna::DumpVectorToString => locking exeption mutex",err)
       
    sRetName << "Pack Size = " << fPack->size() << endl;
    
    for(unsigned int pos=0; pos<fPack->size(); ++pos) {
      
      CDataStructure &dataRAW = fPack->at(pos);

      if(&dataRAW) {
	sRetName << "Event = " << pos << ",  EventTime = " << dataRAW.fEventTime 
		 << ", TriggerChannel = " << dataRAW.fTriggerChannel << ",  GoodEvent = " << dataRAW.fGoodEvent << endl;
	
	sRetName << "Temp 1 = "   << dataRAW.fSlowControl.sTemp1 << ",  Temp 2 = " << dataRAW.fSlowControl.sTemp2
		 << ",  Press = " << dataRAW.fSlowControl.sPress << ",  Efield = " << dataRAW.fSlowControl.sEfVoltage << endl;
	
	sRetName << "  DataPickup  \t  DataNear  \t   DataFar " << endl;

	for(int bin=0; bin<dataRAW.fLocalCopyOfNumberOfADCSampels; ++bin) {

	   sRetName << dataRAW.fDataPickup[bin] << " " << dataRAW.fDataNear[bin] << " " << dataRAW.fDataFar[bin]  << endl;
	}
      } else {
	cerr << "Error in <COnLineAna::DumpVectorToString>: dataRAW is NULL " << flush << endl;
      }
    }
    // unlock 
    err = pthread_mutex_unlock(&fWrapper->fmutex);
    DEBUGINFO("Debuge Info => COnLineAna::DumpVectorToString => unlocking thread mutex",err)
  
  } else {
    cerr << "Error in <COnLineAna::DumpVectorToString>: fPack is NULL " << flush << endl;
  }
  return sRetName.str();
}
