//----------------------------------------------------------------------------
//  Body: CRootFileHandler.cpp, v 1.0                         2007/12/07    //
//                                                                          //
//  Goofie TPC ALICE Program Library                                        //
//  Copyright (C) ALICE collaboration, 2007                                 //
//                                                                          //
//  Last modified in CVS                                                    //
//  Author: D. Antonczyk                                                    //
//----------------------------------------------------------------------------

#include <iostream>
#include <sstream>

#include "TFile.h"
#include "TTree.h"
#include "TBranch.h"

#include "AliGoofiePack.h"
#include "AliGoofieGasResult.h"
#include "AliGoofieRawEvent.h"
          
#include "CDataStructure.h"
#include "CGasAnaResults.h"
#include "SConfigData.h"
#include "CMacroDef.h"
#include "COnLineExeption.h"

#include "CRootFileHandler.h"

// TODO:
//
// TODO optional:
// 1) convert to template class
//

using namespace std;

CRootFileHandler::CRootFileHandler()
{  
   fHandlerForCurrentDir     = NULL;
   fOutputRootFile           = NULL;
   fPackTree                 = NULL;
   fPackBranch               = NULL; 
   fPackObj                  = new AliGoofiePack();          // tree template
   fCurrentFileName          = SConfigData::fRootFileName;   // current file name, template like in the SConfigData::fRootFileName 
   fFileIsReady              = false;

   // root settings
   fCompresionLevel          = 5;
   fBufferSize               = 256000;
   fSplittingValue           = 2;
   fBranchStyle              = 1;

   // storing events before save
   fPackLocalCopy            = new deque<AliGoofieRawEvent>();
   
   // error / exeption handling
   fexeptionStatus  = new COnLineExeption();

   DEBUGINFO("Debuge Info => CRootFileHandler::CRootFileHandler => object created", 0)
}

CRootFileHandler::~CRootFileHandler()
{
   if(fOutputRootFile) {
      fOutputRootFile->Close();
      delete fOutputRootFile;
      fOutputRootFile = NULL;
   }
   if(fPackLocalCopy) {
      fPackLocalCopy->clear();
      delete fPackLocalCopy;
      fPackLocalCopy = NULL;
   }
   // once file is close TTree object is deleted but pointer is not set to NULL
   if(fPackTree) {
      fPackTree = NULL;
   }
   if(fPackObj) {
      delete fPackObj;
      fPackObj = NULL;
   }
   if(fexeptionStatus) {
      delete fexeptionStatus;
      fexeptionStatus = NULL;
   }
   DEBUGINFO("Debuge Info => CRootFileHandler::~CRootFileHandler => object deleted", 0)
}


void CRootFileHandler::Initialize()
{   
   // create local directory for output file
   fHandlerForCurrentDir = gDirectory;
  
   // move to the file directory
   if(fOutputRootFile) {
     fOutputRootFile->cd();
   }

   // create Tree
   if(!fPackTree) {
      fPackTree = new TTree("T","Goofie RAW ROOT tree");
      fPackTree->SetAutoSave(1000000000);  // autosave when 1 Gbyte written
      fPackTree->SetBranchStyle(fBranchStyle);
      if (fSplittingValue)
	 fBufferSize /= 4;
   }
   // get branch
   fPackBranch = fPackTree->Branch("Packs", "AliGoofiePack", &fPackObj, fBufferSize, fSplittingValue);
   if(fPackBranch) {
      fPackBranch->SetAutoDelete(kFALSE);
   } else {
      fexeptionStatus->SetExeptionError(30000000, "<CRootFileHandler::Initialize()>: Initialization of root tree failed");
   }

   // move to previous directory
   fHandlerForCurrentDir->cd();
}


void CRootFileHandler::CopyPack(const deque<CDataStructure> &_fPack)
{
   if(fPackLocalCopy->size() == _fPack.size()) {
      DEBUGINFO("Debuge Info => CRootFileHandler::CopyPack => overwrriting mode used", 0)
      copy(_fPack.begin(), _fPack.end(), fPackLocalCopy->begin());
   } else {
      DEBUGINFO("Debuge Info => CRootFileHandler::CopyPack => copy mode used", 0)
      fPackLocalCopy->clear();
      copy(_fPack.begin(), _fPack.end(), back_inserter(*fPackLocalCopy));
   }
}


void CRootFileHandler::AddPackToTree(const CGasAnaResults *_GasAnaResults, const deque<CDataStructure> &_fPack)
{
   unsigned int err = 0;

   // create local directory for output file
   fHandlerForCurrentDir = gDirectory;
  
   // move to the file directory
   if(fOutputRootFile) {
     fOutputRootFile->cd();
   }
   
   //cast type of objects
   CopyPack(_fPack);
   
   if(fPackObj) {
      fPackObj->BuildEvent(_GasAnaResults, *fPackLocalCopy);
      fPackTree->Fill();
   } else {
      err = 0x30000001;
      fexeptionStatus->SetExeptionError(err, "<CRootFileHandler::AddPackToTree( , )>: Pack object does not exist");
   }

   // move to previous directory
   fHandlerForCurrentDir->cd();

   DEBUGINFO("Debuge Info => CRootFileHandler::AddPackToTree => pack added to tree", err)
}


// test function
void CRootFileHandler::AddPackToTree()
{
   unsigned int err = 0;

   // create local directory for output file
   fHandlerForCurrentDir = gDirectory;
  
   // move to the file directory
   if(fOutputRootFile) {
     fOutputRootFile->cd();
   }

   // dummy  vector
   deque<AliGoofieRawEvent> tmpPack;
   for(int i=0; i<10; ++i) {
      tmpPack.push_back(AliGoofieRawEvent(SConfigData::fNumberOfADCSampels));
   }
   // dummy gas results
   static AliGoofieGasResult _GasAnaResults;
   _GasAnaResults.FillDummyValue();
   
   if(fPackObj) {
      fPackObj->BuildEvent(&_GasAnaResults , tmpPack);
      fPackTree->Fill();
   } else {
      err = 0x30000001;
      fexeptionStatus->SetExeptionError(err, "<CRootFileHandler::AddPackToTree( , )>: Pack object does not exist");
   }

   // move to previous directory
   fHandlerForCurrentDir->cd();
    
   DEBUGINFO("Debuge Info => CRootFileHandler::AddPackToTree() =>  pack added to tree", err)
}


void CRootFileHandler::CreateOutputFile()
{
   // create local directory for output file
   fHandlerForCurrentDir = gDirectory;
   
   // get current Time
   fCurrTime.Set();
      
   ostringstream fileNameAndPath;
   fileNameAndPath << SConfigData::fPathToRootFile << "/" << SConfigData::fRootFileName << "_" << (int)fCurrTime.GetSec() << ".root"; 
   fCurrentFileName = fileNameAndPath.str();
   
#ifdef DEBUG_ONLINE
   cout << "Debuge Info => CRootFileHandler::CreateOutputFile => ROOT file name is " << fCurrentFileName << ", status code = 0" << flush << endl;
#endif
   
   // create output file
   if(!fOutputRootFile) {
      // create file
      fOutputRootFile = new TFile(fCurrentFileName.c_str(), "RECREATE", "Goofie OnLine Data File", fCompresionLevel);
   }    
   // set old directory
   fHandlerForCurrentDir->cd();

   DEBUGINFO("Debuge Info => CRootFileHandler::CreateOutputFile => file has been created", 0)
}


void CRootFileHandler::SaveDataLocally()
{
   // store current directory 
   fHandlerForCurrentDir = gDirectory;
  
   if(fOutputRootFile) {
      // change to the tree directory
      fOutputRootFile->cd();
      // save file
      fOutputRootFile->Write();
      fOutputRootFile->Close();
      delete fOutputRootFile;
      fOutputRootFile = NULL;
      // TTree is deleting while calling TFile->Close()
      fPackTree = NULL;
      
      // go back to previous directory
      fHandlerForCurrentDir->cd();

      DEBUGINFO("Debuge Info => CRootFileHandler::SaveDataLocally => file has been stored", 0)
      
   } else {
      fexeptionStatus->SetExeptionError(0x30000002, "<CRootFileHandler::SaveDataLocally()>: tree is not stored in file");
   }
}

void CRootFileHandler::SendDataToDAQ()
{
   DEBUGINFO("Debuge Info => CRootFileHandler::SendDataToDAQ => not ready", 0)
}
