//----------------------------------------------------------------------------
//  Body: AliGoofieFitFunction.cxx, v 1.0                     2008/08/02    //
//                                                                          //
//  Goofie ALICE Program Library                                            //
//  Copyright (C) ALICE collaboration, 2008                                 //
//                                                                          //
//  Last modified in CVS                                                    //
//  Author: D. Antonczyk                                                    //
//----------------------------------------------------------------------------
#include <iostream>
#include <cmath>

#include "TMinuit.h"
#include "TF1.h"
#include "TH1D.h"
#include "TString.h"

#include "AliGoofiePeakProperties.h"
#include "AliGoofieConfigData.h"

#include "AliGoofieFitFunction.h"

using namespace std;

TH1D      *AliGoofieFitFunction::gfSpectrumToFit = NULL;
Int_t      AliGoofieFitFunction::gfBinStart      = 0;
Int_t      AliGoofieFitFunction::gfBinStop       = 0;
Double_t   AliGoofieFitFunction::gfChi2          = 0.0;


Double_t AliGoofieFitFunction::G4Fit(Double_t *x, Double_t *par)
{
   // par[0] = amplitude
   // par[1] = mpv-(rising edge width)
   // par[2] = rising edge width
   // par[3] = offset

   double res = par[3];
   if(par[1]<=x[0]) {
         if(par[2]!=0)
	 res = par[3] + par[0]*exp(4.0)*pow((x[0]-par[1])/par[2], 4)*exp(-4.0*(x[0]-par[1])/par[2]);
   } else {
      TF1::RejectPoint();
   }
   return res;
}


void AliGoofieFitFunction::fcn(Int_t &npar, Double_t *gin, Double_t &f, Double_t *par, Int_t iflag)
{
   if(gfSpectrumToFit==NULL) return;

   Int_t    nBinUsed = 0;
   Double_t cont     = 0.0;
   Double_t val      = 0.0;
   //calculate - Chi2
   Double_t chi2     = 0.0;

   // overflow and underflow making difference in radii res.
   for (int bin=gfBinStart; bin<=gfBinStop; ++bin) {
      // get bin content
      cont = gfSpectrumToFit->GetBinContent(bin);
      if(cont!=0.0) {
	 val = gfSpectrumToFit->GetXaxis()->GetBinCenter(bin);
	 chi2 +=  pow(cont - G4Fit(&val, par),2)/sqrt(cont);
	 ++nBinUsed;
      }
   }
   f = gfChi2 = chi2;
   // chi2 normalization
   gfChi2 /= nBinUsed+4.0;
}


AliGoofieFitFunction::AliGoofieFitFunction()
{
#ifdef DEBUG_ONLINE
   cout << "Info from <AliGoofieFitFunction()>: constructor called for " << this << flush << endl; 
#endif

   fMinuit = new TMinuit(4);

   fFitResult = new TF1("G4", G4Fit, 0.0, 512.0, 4);
   fFitResult->SetLineColor(2);
   fFitResult->SetLineWidth(1);
   fFitResult->SetNpx(1000);

   fGaussFit = new TF1("fGaussFit", "gaus", 1.0, 3.0);
   fGaussFit->SetLineColor(2);
   fGaussFit->SetLineWidth(1);
   fGaussFit->SetNpx(1000);
}


AliGoofieFitFunction::~AliGoofieFitFunction()
{
#ifdef DEBUG_ONLINE
   cout << "Info from <~AliGoofieFitFunction()>: destructor called for " << this << flush << endl; 
#endif
   if(fMinuit) {
      fMinuit->Delete();
      fMinuit = NULL;
   }
   if(fFitResult) {
      fFitResult->Delete();
      fFitResult = NULL;
   }
   if(fGaussFit) {
      fGaussFit->Delete();
      fGaussFit = NULL;
   }
}


Bool_t AliGoofieFitFunction::MakeFit(TH1D *_histo, AliGoofiePeakProperties *_peakPropert)
{
   // return 0 if nentries equal 0
   if(_histo->GetEntries()<50) {
#ifdef FIT_ERROR
      cerr << "Error in <AliGoofieFitFunction::MakeFit>: not sufficient statistic to performe fit" << flush << endl;
      cerr << "Error in <AliGoofieFitFunction::MakeFit>: this never should happend" << flush << endl;
#endif
      return false;
   }
   // set pointer to data
   gfSpectrumToFit = _histo;

   //set fit range - should be from FirstBin() to LastBin() without under & over flow   
   double StartPos = _peakPropert->fPosition-(0.22*AliGoofieConfigData::fPeakIntegrationRange);
   double StopPos  = _peakPropert->fPosition+(0.78*AliGoofieConfigData::fPeakIntegrationRange);
   gfBinStart = _histo->FindBin(StartPos); 
   gfBinStop  = _histo->FindBin(StopPos);

   //MINUIT loop to get proper errors
   fMinuit->SetFCN(AliGoofieFitFunction::fcn);
   static Double_t arglist[2] = {'\0'};
   Int_t           ierflg = 0;

   arglist[0] = -1; // -1 no info print
   fMinuit->mnexcm("SET PRINT", arglist, 1, ierflg);
   // Supresses Minuit warning messages.
   arglist[0] = 0;
   fMinuit->mnexcm("SET NOW",   arglist, 1, ierflg);

   arglist[0] = 1; // 1-standart, 2-improved 
   fMinuit->mnexcm("SET STR", arglist, 1, ierflg);

   arglist[0] = 1.0;  // for chisquared fits
   fMinuit->mnexcm("SET ERR", arglist, 1, ierflg);

   // set starting values and step size for parameters
   static Double_t step[] = {0.01, 0.001, 0.05, 0.01};

   fMinuit->mnparm(0, "amp",         _peakPropert->fAmplitude,                       step[0], 0, 0, ierflg);
   fMinuit->mnparm(1, "mpv-reWidth", (_peakPropert->fPosition-_peakPropert->fWidth), step[1], 0, 0, ierflg);
   fMinuit->mnparm(2, "reWidth",     _peakPropert->fWidth,                           step[2], 0, 0, ierflg);
   fMinuit->mnparm(3, "offset",      _peakPropert->fOffset,                          step[3], 0, 0, ierflg);

   // ------------------------- Fitting chain ------------------------------
   arglist[0] = 2000;
   arglist[1] = 0.05;
   fMinuit->mnexcm("MINImize", arglist, 2, ierflg);
#ifdef FIT_ERROR
   if(ierflg!=0)
      cout << "Info from <AliGoofieFitFunction::MakeFit>: status of MINImize : " << ierflg << endl;
#endif
   if(ierflg==4) {
      arglist[0] = 1000;
      arglist[1] = 0.01;
      fMinuit->mnexcm("SIMplex", arglist, 2, ierflg);
#ifdef FIT_ERROR
      if(ierflg!=0)
   	 cout << "Info from <AliGoofieFitFunction::MakeFit>: status of SIMplex : " << ierflg << endl;
#endif
   }
   //get fit parameters from MINUIT
   double fitValue[4] = {'\0'};
   double fitErr[4]   = {'\0'};

   fMinuit->GetParameter(0,  fitValue[0],  fitErr[0]);
   fMinuit->GetParameter(1,  fitValue[1],  fitErr[1]);
   fMinuit->GetParameter(2,  fitValue[2],  fitErr[2]);
   fMinuit->GetParameter(3,  fitValue[3],  fitErr[3]);
   // End of MINUIT

   fFitResult->SetParameters(fitValue);
   fFitResult->SetParErrors(fitErr);

   // assign obtained from fit to peak parameter obj.
   _peakPropert->fPosition     = fitValue[1] + fitValue[2];
   _peakPropert->fErrPosition  = hypot(fitErr[1], fitErr[2]);

   _peakPropert->fWidth        = fitValue[2]*2.5;   // only estiamtion that sigma is 2.5 larger then fit with
   _peakPropert->fErrWidth     = fitErr[2]*2.5;

   _peakPropert->fAmplitude    = fitValue[0];
   _peakPropert->fErrAmplitude = fitErr[0];

   _peakPropert->fOffset       = fitValue[3];
   _peakPropert->fErrOffset    = fitErr[3];

   // Set Area
    _peakPropert->fArea = fFitResult->Integral(StartPos, StopPos);

   // Set Chi2/ndof
   _peakPropert->fChi2 = gfChi2;

   return true;
}


TF1 *AliGoofieFitFunction::GetFitResults(AliGoofiePeakProperties *_peakPropert)
{
   double fitValue[4] = {'\0'};
   double fitErr[4]   = {'\0'};

   fitValue[0]   = _peakPropert->fAmplitude;
   fitErr[0]     = _peakPropert->fErrAmplitude;

   fitValue[2]   = _peakPropert->fWidth/2.5;
   fitErr[2]     = _peakPropert->fErrWidth/2.5;

   fitValue[1]   = _peakPropert->fPosition - fitValue[2];
   fitErr[1]     = hypot(_peakPropert->fErrPosition, fitErr[2]);

   fitValue[3]   = _peakPropert->fOffset;
   fitErr[3]     = _peakPropert->fErrOffset;

   fFitResult->SetParameters(fitValue);
   fFitResult->SetParErrors(fitErr);

   TString sFitName = _peakPropert->fName; sFitName += "_FitFunction";
   return (TF1*)fFitResult->Clone(sFitName.Data());
}


bool AliGoofieFitFunction::MakeGaussFit(TH1D *_histo, AliGoofiePeakProperties *_peakPropert)
{
   // return false if number of entries is less then 50 in histogram
   if(_histo->GetEntries()<50) {
      //#ifdef FIT_ERROR
      cerr << "Error in <AliGoofieFitFunction::MakeGaussFit>: not sufficient statistic to performe fit" << flush << endl;
      cerr << "Error in <AliGoofieFitFunction::MakeGaussFit>: this never should happend" << flush << endl;
      //#endif
      return false;
   }
   fGaussFit->SetRange((_peakPropert->fPosition-_peakPropert->fWidth/2.5), (_peakPropert->fPosition+_peakPropert->fWidth/2.5));
   fGaussFit->SetParameters(_peakPropert->fAmplitude,  _peakPropert->fPosition, _peakPropert->fWidth/10.0);
   _histo->Fit(fGaussFit, "RQN");

   _peakPropert->fPosition     = fGaussFit->GetParameter(1);
   _peakPropert->fErrPosition  = fGaussFit->GetParError(1);

   _peakPropert->fAmplitude    = fGaussFit->GetParameter(0);
   _peakPropert->fErrAmplitude = fGaussFit->GetParError(0);

   _peakPropert->fWidth        = fGaussFit->GetParameter(2);
   _peakPropert->fErrWidth     = fGaussFit->GetParError(2);

   _peakPropert->fOffset       = 0.0;
   _peakPropert->fErrOffset    = 0.0;

   _peakPropert->fArea         = 0.0;
   _peakPropert->fChi2         = fGaussFit->GetChisquare()/fGaussFit->GetNDF();
   _peakPropert->fGradient     = 0.0;

   return true;
}


TF1 *AliGoofieFitFunction::GetGaussFitResults(AliGoofiePeakProperties *_peakPropert)
{
   double fitValue[3] = {'\0'};
   double fitErr[3]   = {'\0'};

   fitValue[0]   = _peakPropert->fAmplitude;
   fitErr[0]     = _peakPropert->fErrAmplitude;

   fitValue[1]   = _peakPropert->fPosition;
   fitErr[1]     = _peakPropert->fErrPosition;

   fitValue[2]   = _peakPropert->fWidth;
   fitErr[2]     = _peakPropert->fErrWidth;

   fGaussFit->SetRange(0.0, 500.0);
   fGaussFit->SetParameters(fitValue);
   fGaussFit->SetParErrors(fitErr);

   TString sFitName = _peakPropert->fName; sFitName += "_FitFunction";
   return (TF1*)fGaussFit->Clone(sFitName.Data());
}
