//----------------------------------------------------------------------------
//  Body: CReadNI.cpp,v 1.0                                     2006/01/20  //
//                                                                          //
//  Goofie TPC ALICE Program Library                                        //
//  Copyright (C) ALICE collaboration, 2006                                 //
//                                                                          //
//  Definition of class CReadNI                                             //
//  Last modified in CVS                                                    //
//  Author: D. Antonczyk                                                    //
//----------------------------------------------------------------------------

// TODO
//
//  1) initialization check
//  2) proper mapping, variable to channel no.
//     temporary mapping
//   _  tempA 1 (Near) -> Channel AI0
//   _  tempB 2 (Far)  -> Channel AI1
//      pressure       -> Channel AI2
//      EfieldVoltage  -> Channel AI3
//      HVPickup       -> Channel AI4 not used
//

//STL
#include <fstream>
#include <iostream>
#include <sstream>
#include <cmath>

//ROOT
#include <TString.h>

//
#include "CMacroDef.h"
#include "CCFileParser.h"
#include "COnLineExeption.h"
#include "CDataStructure.h"
#include "SConfigData.h"

#include "CReadNI.h"

using namespace std;

//------------------------ constructor -----------------------------------------//
CReadNI::CReadNI()
{
   // calibration on
   fCalibrationOnOff = true;
   // errors
   ferror            = 0;                   // NI error code 
   // task
   fTaskHandle       = 0;
   // sample buffor
   for(int v=0; v<10; ++v) {
      fReadArray[v]  = 0;
   }
   //common reading parameters
   fMinRangeVolts    = -10.0;                               // min limit (Volts)
   fMaxRangeVolts    = 10.0;                                // max limit (Volts)
   fNumSampsPerChan  = 10;                                  // how many samples to read
   fTimeout          = 10.0;                                // in sec
   fTerminalConfig   = DAQmx_Val_RSE;                       // Referenced single-ended mode
   fFillMode         = DAQmx_Val_GroupByChannel;            // non-interleaved
   fMaxArraySize     = sizeof(fReadArray)/sizeof(float64);  // maximal size of readout array
   fReadArraySize    = 0;	                            // how many sample has been readed
   fexeptionStatus   = new COnLineExeption();
}

//-------------------------- destructor ---------------------------------------//
CReadNI::~CReadNI()
{
   fPhysicalChannelName.clear();
   fAssignedChannelName.clear();
   fChannelValue.clear();
   fChannelUsed.clear();
   fCalibrationValues.clear();

   if(fexeptionStatus) {
      delete fexeptionStatus;
      fexeptionStatus = NULL;
   }
}

//--------------------------- initialization -----------------------------------//
unsigned int CReadNI::InitializeCard(const string _configFile)
{
   // Load setup values
   SetNICalValueFromFile(_configFile);
   // Channel names/states/values
   for(unsigned int _ch=0; _ch<16; ++_ch) {
      stringstream channelName;
      channelName.str("");
      channelName << "Dev1/ai" << _ch;
      fPhysicalChannelName[_ch] = channelName.str();
      //
      channelName.str("");
      channelName << "Dev1_ai" << _ch;
      fAssignedChannelName[_ch] = channelName.str();
      // clear states
      fChannelUsed[_ch] = FALSE;
      // clear values   
      fChannelValue[_ch] = 0.0;
   }
   // check connection with the NI card.
   double channelRes = ReadChannel(0);
   DEBUGINFO("Debuge Info => CReadNI::InitializeCard => returned value from channel AI0  = ", channelRes)
   return ferror;
}

//----------------------------------------------------------------------//
void CReadNI::SetNICalValueFromFile(const string _filename)
{
  try {
    CCFileParser<MapStringDbl_t> cfg(_filename);
    // set values from file
    std::pair<double, double> tempObj;

    string lCalibrationParameters[] = {"TemperatureSensorP0", "TemperatureSensorP1",
				       "TemperatureSensorP0", "TemperatureSensorP1",
				       "PreassureSensorP0",   "PreassureSensorP1",
				       "FieldHVP0",           "FieldHVP1",
				       "PickupVoltageP0",     "PickupVoltageP1"};
    
    for(unsigned int size=0; size<(sizeof(lCalibrationParameters)/sizeof(string)/2); ++size) {
       for(unsigned int pairPos=0; pairPos<2; ++pairPos) {
          int index = pairPos + 2*size;
	  if(pairPos==0) 
	     tempObj.first  = cfg.get_value(lCalibrationParameters[index]);
	  else 
	     tempObj.second = cfg.get_value(lCalibrationParameters[index]);
       }
       fCalibrationValues.push_back(tempObj);
    }
  } catch(exception& e) {
    string errorInfp = "<CReadNI::SetNICalValueFromFile()>:";
    errorInfp +=  e.what();
    fexeptionStatus->SetExeptionError(0x0000001, errorInfp.c_str());
  }
}

void CReadNI::ResetUsedChannel()
{
  for(unsigned int _ch=0; _ch<fChannelUsed.size(); ++_ch) {
    fChannelValue[_ch] = false;
  }
}

//----------------------------------------------------------------------//
unsigned int CReadNI::GetSCData(CDataStructure *_eventStr, bool _update)
{
   if(_update) {
      ReadDataForGOOFIE();
   }
   // set new values
   _eventStr->fSlowControl.sTemp1     = fChannelValue[0];
   _eventStr->fSlowControl.sTemp2     = fChannelValue[1];
   _eventStr->fSlowControl.sPress     = fChannelValue[2];
   _eventStr->fSlowControl.sEfVoltage = fChannelValue[3];
   _eventStr->fSlowControl.sHVPickup  = fChannelValue[4];

   return ferror;
}

//----------------------------------------------------------------------//
double CReadNI::ReadChannel(unsigned int _chNr)
{
   try {
      // used channel
      fChannelUsed[_chNr] = TRUE;
      // clean
      fChannelValue[_chNr] = 0.0;
      // create task
      ferror = DAQmxBaseCreateTask("", &fTaskHandle);
      if(DAQmxFailed(ferror)) throw (int32)-1;
      // create channel
      ferror = DAQmxBaseCreateAIVoltageChan(fTaskHandle, fPhysicalChannelName[_chNr].c_str(), fAssignedChannelName[_chNr].c_str(),
					    fTerminalConfig, fMinRangeVolts, fMaxRangeVolts, DAQmx_Val_Volts, NULL);
      if(DAQmxFailed(ferror)) throw (int32)-2;
      // start task
      ferror =  DAQmxBaseStartTask(fTaskHandle);
      if(DAQmxFailed(ferror)) throw (int32)-3;
      // read values
      ferror =  DAQmxBaseReadAnalogF64(fTaskHandle, fNumSampsPerChan, fTimeout, fFillMode, fReadArray, fMaxArraySize, &fReadArraySize, NULL);
      if(DAQmxFailed(ferror)) throw (int32)-4;

   } catch (int32 _error) {
      // error info
      char errBuff[2048]= {'\0'};
      if( DAQmxFailed(_error) ) {
	 DAQmxBaseGetExtendedErrorInfo(errBuff, 2048);
	 stringstream ss;
	 ss << "<CReadNI::ReadChannel()>: " << errBuff;
	 fexeptionStatus->SetExeptionError(ferror, ss.str().c_str());
      }
   }
   // stop/ clear
   if(fTaskHandle != 0) {
      DAQmxBaseStopTask(fTaskHandle);
      DAQmxBaseClearTask(fTaskHandle);
   }
   // calculate average value
   if(fReadArraySize>0) {
     for(int samples=0; samples<fReadArraySize; ++samples) {
       //cout << "test : val ch[" << samples << "] = " << fReadArray[samples] << endl;
       fChannelValue[_chNr] += (double)fReadArray[samples];
     }
     fChannelValue[_chNr] /= fReadArraySize;
   } else {
     fexeptionStatus->SetExeptionError(ferror, "<CReadNI::ReadChannel()>: read-out of NI card failed");
   }
   DEBUGINFO("Debuge Info => CReadNI::ReadChannel => read-out from NI card performed correctly", ferror)

   return fChannelValue[_chNr];
}

//----------------------------------------------------------------------//
void  CReadNI::ReadDataForGOOFIE()
{
   ReadChannel(0);  // Temp A // Near
   ReadChannel(1);  // Temp B // Far
   ReadChannel(2);  // Pressure
   ReadChannel(3);  // HV
   ReadChannel(4);  // Pickup HV

   if(fCalibrationOnOff) {
      Calibrate();
   }
}

//----------------------------------------------------------------------//
void CReadNI::Calibrate()
{
   //calibrate
   for(unsigned int ch=0; ch<fCalibrationValues.size(); ++ch) {
      fChannelValue[ch] = (GetChannelValue(ch)*fCalibrationValues[ch].second) + fCalibrationValues[ch].first;
      // test
      //cout << "channel[" << ch << "] P0="<< fCalibrationValues[ch].first << " P1=" << fCalibrationValues[ch].second << endl;
   }
   DEBUGINFO("Debuge Info => CReadNI::Calibrate => incoming data has been calibrated", 1)
}


//----------------------------------------------------------------------//
//-------                        TEST FUNCTIONS                   ------//
//----------------------------------------------------------------------//

void CReadNI::ShowAcq()
{
   for(unsigned int _ch=0; _ch<fChannelUsed.size(); ++_ch) {
      if(fChannelUsed[_ch]) {
	 cout<< " channel[" << _ch << "]=" << fChannelValue[_ch] << ",\t";
      }
   }
   cout << endl;
}
//----------------------------------------------------------------------//

void CReadNI::SaveDataInFile(ofstream *fout)
{
   //take the data stream and store it in a given file
   if (!fout) {
      cerr << " no output file defined " << endl;
   } else {
      for(unsigned int _ch=0; _ch<fChannelUsed.size(); ++_ch) {
	 if(fChannelUsed[_ch]) {
	    *fout << " channel[" << _ch << "]=" << fChannelValue[_ch] << ",\t";
	 }
      }
   }
   *fout << endl;
}
//----------------------------------------------------------------------//

string CReadNI::GetChannelsValues()
{
   stringstream channelName;
   channelName.str("");

   for(unsigned int _ch=0; _ch<fChannelUsed.size(); ++_ch) {
      if(fChannelUsed[_ch]) {
	 channelName << " ch[" << _ch << "]=" << fChannelValue[_ch] << ",\t";
      }
   }
   return channelName.str();
}
