/************************************************************************/
/* file:    altro_driver_26.c						*/
/*  author:  Markus Joos, CERN-PH-ESS 					*/
/*									*/
/* This diver based on: USB Skeleton driver - 1.1			*/
/*									*/
/*  This driver works for kernels from 2.6 onwards			*/
/*									*/
/*  Open issues (according to C. H. Christiansen, may be obsolete):	*/
/*									*/
/*  1) The memory read functions do not work because the IN		*/
/*  transaction is executed before the OUT transaction 			*/
/*									*/
/*  2) If one repeats the same memory write two times			*/
/*  (e.g. IMEM_Write) the first transaction has an ACK,			*/   
/*  the second a NYet, the third URB does not complete and the		*/
/*  fourth creates the error mentioned in issue 1) above  		*/
/*									*/
/*  3) The memory read/write calls in the U2F library need an offset	*/
/*  parameter relative to the respective memory				*/
/************************************************************************/


#include <linux/autoconf.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/smp_lock.h>
#include <linux/completion.h>
#include <linux/types.h>
#include <linux/usb.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
#include "altro_common.h"


#ifdef ALTRO_KERNEL_DEBUG
  #define kdebug(x) {if (debug) printk x;}
#else
  #define kdebug(x)
#endif

#define kerror(x) {if (errorlog) printk x;}

#define DRIVER_VERSION "v2.0"
#define DRIVER_DESC    "USB driver for the ALICE ALTRO card"

#define USB_ALTRO_VENDOR_ID_CY 0x04b4 
#define USB_ALTRO_GOFFIE_ID_CY 0x1003

#define USB_ALTRO_VENDOR_ID	 0x1556
#define USB_ALTRO_PRODUCT_ID 	 0x0350
#define USB_ALTRO_U2F_ID         0x15e
#define USB_ALTRO_GOFFIE_ID      0x15f
#define USB_ALTRO_MPT_ID         0x163
#ifndef USB_ALTRO_MINOR_BASE
# define USB_ALTRO_MINOR_BASE	 192
#endif

/* Turn Jiffies into milliseconds.  Since 2.6.12, the time-out */
/* parameter of usb_bulk_msg should be in seconds, not jiffies. */
#if HZ == 1000
# define JIFFIES_TO_MSECS(x) (x)
# define MSECS_TO_JIFFIES(x) (x)
#elif HZ == 100
# define JIFFIES_TO_MSECS(x) ((x) * 10)
# define MSECS_TO_JIFFIES(x) (((x) + 9) / 10)
#else
# define JIFFIES_TO_MSECS(x) (((x) * 1000) / HZ)
# define MSECS_TO_JIFFIES(x) (((x) * HZ + 999) / 1000)
#endif
#if  LINUX_VERSION_CODE > KERNEL_VERSION(2,6,11) 
# if HZ == 1000
#  define TIMEOUT(x) (x)
# elif HZ == 100
#  define TIMEOUT(x) ((x) * 10)
# else 
#  define TIMEOUT(x) (((x)  * 1000) / HZ)
# endif
#else 
# define  TIMEOUT(x) (x)
#endif

static int debug = 1, errorlog = 1, ddump = 0;

module_param (ddump, int, S_IRUGO | S_IWUSR);  /*MJ: for 2.6 p37*/
MODULE_PARM_DESC(ddump, "1 = enable data dumping   0 = disable data dumping");

module_param (debug, int, S_IRUGO | S_IWUSR);  
MODULE_PARM_DESC(debug, "1 = enable debugging   0 = disable debugging"); 

module_param (errorlog, int, S_IRUGO | S_IWUSR);  
MODULE_PARM_DESC(errorlog, "1 = enable error logging   0 = disable error logging");

MODULE_AUTHOR("Markus Joos, CERN/PH");
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("Private: Contact markus.joos@cern.ch");

static struct usb_device_id altro_table [] = 
{
  { USB_DEVICE(USB_ALTRO_VENDOR_ID, USB_ALTRO_PRODUCT_ID), driver_info: (u_long)"altro_by_vendor (invalid device ID)" },
  { USB_DEVICE(USB_ALTRO_VENDOR_ID_CY, USB_ALTRO_GOFFIE_ID_CY), driver_info: (u_long)"goofie_by_vendor (Cypress device ID)" },
  { USB_DEVICE(USB_ALTRO_VENDOR_ID, USB_ALTRO_U2F_ID), driver_info: (u_long)"ALICE U2F" },
  { USB_DEVICE(USB_ALTRO_VENDOR_ID, USB_ALTRO_GOFFIE_ID), driver_info: (u_long)"ALICE GOFFIE" },
  { USB_DEVICE(USB_ALTRO_VENDOR_ID, USB_ALTRO_MPT_ID), driver_info: (u_long)"ALICE MPT" },
  { } /* Terminating entry */
};

MODULE_DEVICE_TABLE (usb, altro_table);

struct altro_device 
{
  
  struct usb_device* udev;		/* save off the usb device pointer */ 
  struct usb_interface*	interface;	/* the interface for this device */ 
  u_char minor;		  		/* the starting minor number for this device */ 
  u_char	num_ports;	  	/* the number of ports this device has */ 
  char num_interrupt_in;	  	/* number of interrupt in endpoints we have */ 
  char num_bulk_in;		  	/* number of bulk in endpoints we have */
  char num_bulk_out;		  	/* number of bulk out endpoints we have */ 
  u_char* bulk_in_buffer;		/* the buffer to receive data */ 
  size_t bulk_in_size;		  	/* the size of the receive buffer */ 
  struct urb* read_urb;		  	/* the urb used to recieve data */ 
  __u8 bulk_in_endpointAddr;	  	/* the address of the bulk in endpoint */
  u_char* bulk_out_buffer;		/* the buffer to send data */ 
  size_t bulk_out_size;		 	/* the size of the send buffer */ 
  struct urb* write_urb;	  	/* the urb used to send data */ 
  __u8 bulk_out_endpointAddr;	  	/* the address of the bulk out endpoint */ 
  atomic_t write_busy;		  	/* true iff write urb is busy */ 
  struct completion write_finished;	/* wait for the write to finish */ 
  int open;			  	/* if the port is open or not */ 
  int present;		  		/* if the device is not disconnected */ 
  struct semaphore sem;		 	/* locks this structure */ 
  char* name;
};

static u_int n_our_devices = 0;
static struct altro_device* our_devices[10] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

static DECLARE_MUTEX (disconnect_sem);

struct altro_proc_data_t
{
  char name[10];
  char value[100];
};

static struct proc_dir_entry* altro_proc_file;
static struct altro_proc_data_t altro_proc_data;


/* Prototypes*/
static int altro_ioctl(struct inode *inode, struct file *file, u_int cmd, u_long arg);
static int altro_open(struct inode *inode, struct file *file);
static int altro_release(struct inode *inode, struct file *file);
static int altro_probe(struct usb_interface *interface, const struct usb_device_id *id);
static int altro_proc_write(struct file *file, const char *buffer, u_long count, void *data);
static int altro_proc_read(char *page, char **start, off_t off, int count, int *eof, void *data);
static void altro_disconnect(struct usb_interface *interface);
static void altro_write_bulk_callback(struct urb *urb, struct pt_regs *regs);
static void add_our_device(struct altro_device* dev); 
static void remove_our_device(struct altro_device* dev); 
static ssize_t altro_read(struct file *file, char *buffer, size_t count, loff_t *ppos);
static ssize_t altro_write(struct file *file, const char *buffer, size_t count, loff_t *ppos);


static struct file_operations altro_fops = 
{
  .owner =	THIS_MODULE,
  .ioctl =	altro_ioctl,
  .open =	altro_open,
  .release =	altro_release,
  .read =	altro_read,
  .write =	altro_write
};

static struct usb_class_driver altro_class = 
{
  .name = "altro%d",
  .fops = &altro_fops,
  // 28/08/08 Dariusz Antonczyk, necessary for kernel 2.6.22.1	
  //.mode = S_IFCHR | S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH |  S_IWOTH,
  .minor_base =	USB_ALTRO_MINOR_BASE,
};

static struct usb_driver altro_driver = 
{
  //.owner =	THIS_MODULE,  // 28/08/08 Dariusz Antonczyk, necessary for kernel 2.6.22.1
  .name =	"altro",
  .probe =	altro_probe,
  .disconnect =	altro_disconnect,
  .id_table =	altro_table,
};


/**************************************************/
static void add_our_device(struct altro_device* dev) 
/**************************************************/
{
  kdebug(("altro_driver(add_our_device): Function called\n"));

  if (n_our_devices >= 10) 
  {
    kerror(("altro_driver(add_our_device): Cannot add the 11th device to internal table\n"));
    return;
  }
  our_devices[n_our_devices] = dev;
  kdebug(("altro_driver(add_our_device): Added device %s at %d\n", dev->name, n_our_devices));
  n_our_devices++;
  kdebug(("altro_driver(add_our_device): Function done\n"));
}


/*****************************************************/
static void remove_our_device(struct altro_device* dev) 
/*****************************************************/
{
  u_int i = 0;

  kdebug(("altro_driver(remove_our_device): Function called\n"));
  for (i = 0; i < n_our_devices; i++) 
  {
    if (dev == our_devices[i]) 
    {
      our_devices[i] = 0;
      n_our_devices--;
      kdebug(("altro_driver(remove_our_device): Removed device %s at %d\n", dev->name, i));
      break;
    }
  }
  
  for (; i < n_our_devices + 1; i++) 
    our_devices[i] = our_devices[i + 1];
    
  if (i >= n_our_devices + 1)
    our_devices[i] = 0;
  kdebug(("altro_driver(remove_our_device): Function done\n"));
}


/*******************************************************/
static inline void altro_delete(struct altro_device *dev)
/*******************************************************/
{
  kdebug(("altro_driver(altro_delete): Function called\n"));
  kfree (dev->bulk_in_buffer);
  usb_buffer_free(dev->udev, dev->bulk_out_size, dev->bulk_out_buffer, dev->write_urb->transfer_dma);
  usb_free_urb(dev->write_urb);
  kfree(dev);
  kdebug(("altro_driver(altro_delete): Function done\n"));
}


/***********************************************************/
static int altro_open(struct inode *inode, struct file *file)
/***********************************************************/
{
  struct altro_device* dev = NULL;
  struct usb_interface* interface;
  int subminor, retval = 0;
  
  kdebug(("altro_driver(altro_open): Function called\n"));

  /* the below is if we have more than one USB device.*/
  subminor = iminor(inode);
  kdebug(("altro_driver(altro_open): minor number = %d\n", subminor));

  down(&disconnect_sem);  /* prevent disconnects */

  interface = usb_find_interface (&altro_driver, subminor);
  if (!interface) 
  {
    kerror(("altro_drv(altro_open): failed to get minor number %d\n", subminor));
    retval = -ENODEV;
    goto exit_no_device;
  }

  dev = usb_get_intfdata(interface);
  if (!dev) 
  {
    kerror(("altro_drv(altro_open): Error from call to usb_get_intfdata\n"));
    retval = -ENODEV;
    goto exit_no_device;
  }

  down(&dev->sem);            /* lock this device */
  ++dev->open;                /* increment our usage count for the driver */
  file->private_data = dev;   /* save our object in the file's private structure */
  up (&dev->sem);             /* unlock this device */

 exit_no_device:
  up (&disconnect_sem);

  kdebug(("altro_driver(altro_open): Function done\n"));
  return retval;
}


/**************************************************************/
static int altro_release(struct inode *inode, struct file *file)
/**************************************************************/
{
  int retval = 0;
  struct altro_device* dev;

  kdebug(("altro_driver(altro_release): Function called\n"));
  dev = (struct altro_device *)file->private_data;
  if (dev == NULL) 
  {
    kerror(("altro_driver(altro_release): dev object is NULL\n"));
    return -ENODEV;
  }
  kdebug(("altro_driver(altro_release): minor %d\n", dev->minor));

  down (&dev->sem);  /* lock our device */
  if (dev->open <= 0) 
  {
    kdebug(("altro_driver(altro_release): device not opened\n"));
    retval = -ENODEV;
    goto exit_not_opened;
  }

  /* wait for any bulk writes that might be going on to finish up */
  if (atomic_read (&dev->write_busy))
    wait_for_completion (&dev->write_finished);

  --dev->open;

  if (!dev->present && !dev->open) 
  {
    up (&dev->sem);
    altro_delete (dev);
    kdebug(("altro_driver(altro_release): The device was unplugged before the file was released\n"));
    kdebug(("altro_driver(altro_release): Function done\n"));
    return 0;
  }

 exit_not_opened:
  up (&dev->sem);
  kdebug(("altro_driver(altro_release): Function done\n"));
  return retval;
}


/************************************************************************************/
static ssize_t altro_read(struct file *file, char *buffer, size_t count, loff_t *ppos)
/************************************************************************************/
{
  struct altro_device *dev;
  int retval = 0;

  kdebug(("altro_driver(altro_read): Function called\n"));
  dev = (struct altro_device *)file->private_data;
  kdebug(("altro_driver(altro_read): minor %d, count = %d\n", dev->minor, count));

  down (&dev->sem);  /* lock this object */

  /* verify that the device wasn't unplugged */
  if (!dev->present) 
  {
    kerror(("altro_driver(altro_read): Error: Device was unplugged!\n"));
    up (&dev->sem);
    return -ENODEV;
  }

  kdebug(("altro_driver(altro_read): Function called for end-point 0x%08x. bulk_in_size = %d\n", dev->bulk_in_endpointAddr, dev->bulk_in_size));
  
  /* do a blocking bulk read to get data from the device */ 
  retval = usb_bulk_msg(dev->udev,
			usb_rcvbulkpipe (dev->udev, dev->bulk_in_endpointAddr),
			dev->bulk_in_buffer,
			min (dev->bulk_in_size, count),
			&count, HZ*10);

  /* if the read was successful, copy the data to userspace */
  if (!retval) 
  {
    if (copy_to_user (buffer, dev->bulk_in_buffer, count)) 
    {
      kerror(("altro_driver(altro_read): error from copy_to_user\n"));
      retval = -EFAULT;
    }
    else
      retval = count;
  }
  else 
    kdebug(("altro_driver(altro_read): No data recivied\n"));
  
  /* unlock the device */
  up (&dev->sem);
  kdebug(("altro_driver(altro_read): Function done\n"));
  return retval;
}


/*******************************************************************************************/
static ssize_t altro_write(struct file *file, const char *buffer, size_t count, loff_t *ppos) 
/*******************************************************************************************/
{
  struct altro_device* dev;
  ssize_t bytes_written = 0;
  int retval = 0;
 
  kdebug(("altro_driver(altro_write): Function called\n"));
  dev = (struct altro_device *)file->private_data;
  kdebug(("altro_driver(altro_write): minor %d, count = %d\n", dev->minor, count)); 

  down (&dev->sem);   /* lock this object */ 

  /* verify that the device wasn't unplugged */ 
  if (!dev->present) 
  { 
    kerror(("altro_driver(altro_write): Error: Device was unplugged!\n")); 
    retval = -ENODEV; 
    goto exit; 
  }

  kdebug(("altro_driver(altro_write): Function called for end-point 0x%08x. bulk_in_size = %d\n", dev->bulk_in_endpointAddr, dev->bulk_in_size));

  /* verify that we actually have some data to write */ 
  if (count == 0) 
  { 
    kerror(("altro_driver(altro_write): write request of 0 bytes\n")); 
    goto exit; 
  } 

  /* wait for a previous write to finish up; we don't use a timeout
   * and so a nonresponsive device can delay us indefinitely. */
  if (atomic_read (&dev->write_busy))
    wait_for_completion (&dev->write_finished);

  /* we can only write as much as our buffer will hold */
  bytes_written = min (dev->bulk_out_size, count);

  /* copy the data from userspace into our transfer buffer;
   * this is the only copy required.   */
  if (copy_from_user(dev->write_urb->transfer_buffer, buffer, bytes_written)) 
  {
    kerror(("altro_driver(altro_write): error from copy_from_user\n"));
    retval = -EFAULT;
    goto exit;
  }

  /* this urb was already set up, except for this write size */
  dev->write_urb->transfer_buffer_length = bytes_written;

  /* send the data out the bulk port */
  /* a character device write uses GFP_KERNEL, unless a spinlock is held */
  init_completion (&dev->write_finished);
  atomic_set (&dev->write_busy, 1);
  retval = usb_submit_urb(dev->write_urb, GFP_KERNEL);
  if (retval) 
  {
    atomic_set (&dev->write_busy, 0);
    kerror(("altro_driver(altro_write): Failed submitting write urb, error = %d\n", retval)); 
    kerror(("altro_driver(altro_write): No data sent\n"));
  } 
  else 
  {
    kdebug(("altro_driver(altro_write): %d bytes sent\n", bytes_written));
    retval = bytes_written; 
  }

 exit:
  /* unlock the device */
  up (&dev->sem);

  kdebug(("altro_driver(altro_write): Function done\n"));
  return retval;
}


/* Call-back for the ioctl system call.  This driver understands three commands	*/
/*    SEND_DATA									*/
/*      Send data to the ALTRO device. In this case, the input			*/
/*      parameter "arg" is taken to be a pointer to an 				*/
/*      altro_bulk_out_t variable allocated by the user. 			*/
/*    GET_DATA 									*/
/*      Get data from the ALTRO device. In this case, the input			*/
/*      parameter "arg" is taken to be a pointer to an 				*/
/*      altro_bulk_in_t variable allocated by the user. 			*/
/*    CONTROL 									*/
/*      Control message to the ALTRO device. In this case, the input		*/
/*      parameter "arg" is taken to be a pointer to an 				*/
/*      altro_control_t variable allocated by the user. 			*/
 
/***********************************************************************************/ 
static int altro_ioctl(struct inode *inode, struct file *file, u_int cmd, u_long arg)
/***********************************************************************************/ 
{
  struct altro_device *dev;
  int retval = 0, pipe;
  u_int loop;
  
  kdebug(("altro_driver(altro_ioctl): function called for %s\n", (cmd == GET_DATA ? "GET_DATA" : (cmd == SEND_DATA ? "SEND_DATA" : (cmd == CONTROL ? "CONTROL" : "unknown")))));
  
  /* Get the device */
  dev = (struct altro_device*)file->private_data;
  
  down (&dev->sem);  /* lock this object */

  /* verify that the device wasn't unplugged */
  if (!dev->present) 
  {
    kdebug(("Error: Device was unplugged!\n"));
    retval = -ALTRO_ENODEV;
    goto ioctl_return;
  }
  kdebug(("altro_driver(altro_ioctl): minor %d, cmd 0x%.4x, arg %ld\n", dev->minor, cmd, arg));

  /* fill in your device specific stuff here */
  switch (cmd) 
  {
    case GET_DATA: 
    {
      altro_bulk_in_t params;
      kdebug(("altro_driver(altro_ioctl): GET_DATA called for endpoint 0x%08x\n", dev->bulk_in_endpointAddr));
      kdebug(("altro_driver(altro_ioctl): in bulk size is %d\n", dev->bulk_in_size));

      /* Set up our URB */
      pipe = usb_rcvbulkpipe(dev->udev, dev->bulk_in_endpointAddr);
      kdebug(("altro_driver(altro_ioctl): writing on pipe %d\n", pipe));

      /* do an immediate bulk read to get data from the device */
      retval = usb_bulk_msg(dev->udev, pipe, &(params.data[0]), dev->bulk_in_size, &(params.nbytes), TIMEOUT(HZ * 10));
      kdebug(("altro_driver(altro_ioctl): USB bulk message returns %d\n", retval));
      if (retval) 
      {
	kdebug(("altro_driver(altro_ioctl): no data recieved!\n"));
	params.nbytes = 0;
	retval = -ALTRO_NODATA;
      }
      else 
	kdebug(("altro_driver(altro_ioctl): %d bytes recieved\n", params.nbytes));

      if (ddump)
      {
	kdebug(("altro_driver(altro_ioctl): Now dumping newly received data.......\n"));
        for (loop = 0; loop < params.nbytes; loop++)
	  kdebug(("altro_driver(altro_ioctl): Byte #%d = 0x%02x\n", loop, params.data[loop]));
      } 

      /* Copy data to user */
      if (copy_to_user((void*)arg, &params, sizeof(altro_bulk_in_t)) != 0)
      {
	kerror(("altro_driver(altro_ioctl): error when copying to user\n"));
	retval = -ALTRO_EFAULT;
      }
    }
    break;
    
    case SEND_DATA: 
    {
      altro_bulk_out_t params;
      int maxp, ndone;
      kdebug(("altro_driver(altro_ioctl): SEND_DATA called for endpoint 0x%08x\n", dev->bulk_out_endpointAddr));
      kdebug(("altro_driver(altro_ioctl): in bulk size is %d\n", dev->bulk_out_size));

      /* Copy data from user */
      if (copy_from_user(&params, (void*)arg, sizeof(altro_bulk_out_t))!= 0)
      {
	kerror(("altro_driver(altro_ioctl): error when copying from user\n"));
	retval = -ALTRO_EFAULT;
	goto ioctl_return;
      }

      /* Set up our URB */
      pipe = usb_sndbulkpipe(dev->udev, dev->bulk_out_endpointAddr); 
      kdebug(("altro_driver(altro_ioctl): reading from pipe %d\n", pipe));
      maxp = usb_maxpacket(dev->udev, pipe, usb_pipeout(pipe));
      if (maxp > dev->bulk_out_size) 
        maxp = dev->bulk_out_size;
      kdebug(("altro_driver(altro_ioctl): it is possible to send %d bytes\n", maxp));

      /* Check if we would write more data than possible */
      if (params.nbytes > maxp) 
      {
	kerror(("altro_driver(altro_ioctl): cannot write %d bytes to pipe of %d bytes\n", params.nbytes, maxp));
	retval = -ALTRO_EFAULT;
	goto ioctl_return;
      }

      kdebug(("altro_driver(altro_ioctl): trying to send %d bytes\n", params.nbytes));
      
      if (ddump)
      {
	kdebug(("altro_driver(altro_ioctl): Dumping data berfore sending it.......\n"));
        for (loop = 0; loop < params.nbytes; loop++)
	  kdebug(("altro_driver(altro_ioctl): Byte #%d = 0x%02x\n", loop, params.data[loop]));
      } 
      
      /* do an immediate bulk write to send data to the device */
      retval = usb_bulk_msg(dev->udev, pipe, &(params.data[0]), params.nbytes, &ndone, TIMEOUT(HZ * 10));
      kdebug(("altro_driver(altro_ioctl): USB bulk message returns %d\n",retval));
      if (retval) 
      {
	kdebug(("altro_driver(altro_ioctl): no data sent!\n"));
	params.nbytes = 0;
	retval = -ALTRO_NODATA;
      }
      else 
	kdebug(("altro_driver(altro_ioctl): %d bytes sent\n", params.nbytes));
    } 
    break;

    case CONTROL: 
    {
      altro_control_t params;
      kdebug(("altro_driver(altro_ioctl): CONTROL called\n"));

      /* Copy data from user */
      if (copy_from_user(&params, (void*)arg, sizeof(altro_control_t)) != 0)
      {
	kerror(("altro_driver(altro_ioctl): error when copying from user\n"));
	retval = -ALTRO_EFAULT;
	goto ioctl_return;
      }

      kdebug(("altro_driver(altro_ioctl): CONTROL: bRequest      = %d\n", params.bRequest));   
      kdebug(("altro_driver(altro_ioctl): CONTROL: bmRequestType = %d\n", params.bmRequestType));
      kdebug(("altro_driver(altro_ioctl): CONTROL: wValue        = %d\n", params.wValue));
      kdebug(("altro_driver(altro_ioctl): CONTROL: wIndex        = %d\n", params.wIndex));
      kdebug(("altro_driver(altro_ioctl): CONTROL: data[0]       = %d\n", params.data[0]));
      kdebug(("altro_driver(altro_ioctl): CONTROL: wLength       = %d\n", params.wLength));

      retval = usb_control_msg(dev->udev, 
			       usb_rcvctrlpipe(dev->udev, 0),
			       params.bRequest, 
			       params.bmRequestType,
			       params.wValue,
			       params.wIndex,
			       &(params.data[0]),  
			       params.wLength, TIMEOUT(HZ));	  
      if (retval) 
      {
	kerror(("altro_driver(altro_ioctl): error when sending control word: %d\n", retval));
	retval = -ALTRO_EFAULT;
      }
    }
    break;

    default:
      retval = -ALTRO_ENOTTY;
    break;
  }
  
 ioctl_return:  
  up (&dev->sem);  /* unlock the device */

  kdebug(("altro_driver(altro_IOCTL): Function done\n"));
  return retval; 
}


/***************************************************************************/
static void altro_write_bulk_callback (struct urb *urb, struct pt_regs *regs)
/***************************************************************************/
{
  struct altro_device *dev = (struct altro_device *)urb->context;

  kdebug(("altro_driver(altro_write_bulk_callback): Function called\n"));
  kdebug(("altro_driver(altro_write_bulk_callback): minor %d\n", dev->minor));

  /* sync/async unlink faults aren't errors */
  if (urb->status && !(urb->status == -ENOENT || urb->status == -ECONNRESET)) 
  {
    kdebug(("altro_driver(altro_write_bulk_callback): nonzero write bulk status received: %d\n", urb->status));
  }
  
  /* notify anyone waiting that the write has finished */
  atomic_set (&dev->write_busy, 0);
  complete (&dev->write_finished);
  kdebug(("altro_driver(altro_write_bulk_callback): Function done\n"));
}


/*************************************************************************************/
static int altro_probe(struct usb_interface *interface, const struct usb_device_id *id)
/*************************************************************************************/
{
  int i, retval = -ENOMEM;
  struct usb_device *udev                  = interface_to_usbdev(interface);
  struct altro_device *dev                 = NULL;
  struct usb_host_interface *iface_desc    = 0;
  struct usb_endpoint_descriptor *endpoint = 0;
  size_t buffer_size;

  kdebug(("altro_driver(altro_probe): Function called\n"));
  kdebug(("altro_driver(altro_probe): checking if we want to handle this device %d:%d (%d:%d)\n", udev->descriptor.idVendor, udev->descriptor.idProduct, USB_ALTRO_VENDOR_ID, USB_ALTRO_PRODUCT_ID));

  /* See if the device offered us matches what we can accept */
  if ((udev->descriptor.idVendor != USB_ALTRO_VENDOR_ID) && (udev->descriptor.idVendor != USB_ALTRO_VENDOR_ID_CY)) 
  {
    kerror(("altro_driver(altro_probe): This driver does not handle devices with VENDOR_ID = %d\n", udev->descriptor.idVendor));
    return -ENODEV;
  }
  
  /* See if the device offered us matches what we can accept */
  switch (udev->descriptor.idProduct) 
  {
    case USB_ALTRO_U2F_ID:    break;
    case USB_ALTRO_GOFFIE_ID: break; 
    case USB_ALTRO_GOFFIE_ID_CY: break; 
    case USB_ALTRO_MPT_ID:    break;
    default: 
      kerror(("altro_driver(altro_probe): This driver does not handle devices with PRODUCT_ID = %d\n", udev->descriptor.idProduct));
      return -ENODEV;
  }
  
  kdebug(("altro_driver(altro_probe): got a %s card (%X,%X)\n", (char*)(id->driver_info), udev->descriptor.idVendor, udev->descriptor.idProduct));
  
  /* allocate memory for our device state and initialize it */
  dev = kmalloc (sizeof(struct altro_device), GFP_KERNEL);
  if (dev == NULL) 
  {
    kerror(("altro_driver(altro_probe): Out of memory\n"));
    return -ENOMEM;
  }
  memset (dev, 0x00, sizeof (*dev));

  dev->name      = (char*)id->driver_info;
  init_MUTEX (&dev->sem);
  dev->udev      = udev;
  dev->interface = interface;

  /* set up the endpoint information */
  iface_desc = &interface->altsetting[0];
  if (iface_desc->desc.bNumEndpoints != 2) 
  {
    kerror(("altro_driver(altro_probe): Error: number of endpoints is %d (should be 2)\n", iface_desc->desc.bNumEndpoints));
    goto error;
  }
  
  /* check out the endpoints */
  /* use only the first bulk-in and bulk-out endpoints */
  for (i = 0; i < iface_desc->desc.bNumEndpoints; ++i) 
  {
    endpoint = &iface_desc->endpoint[i].desc;

    kdebug(("altro_driver(altro_probe): Processing end point %d\n", i));
    if ((endpoint->bEndpointAddress & USB_DIR_IN) && ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) == USB_ENDPOINT_XFER_BULK)) 
    {
      /* we found a bulk in endpoint */
      if (!dev->bulk_in_endpointAddr) 
      {
	kdebug(("altro_driver(altro_probe): we found the first bulk in endpoint\n"));
	buffer_size               = endpoint->wMaxPacketSize;
	dev->bulk_in_size         = buffer_size;
	dev->bulk_in_endpointAddr = endpoint->bEndpointAddress;
	dev->bulk_in_buffer       = kmalloc(buffer_size, GFP_KERNEL);
	kdebug(("altro_driver(altro_probe): bulk in endpoint address: %d\n", dev->bulk_in_endpointAddr));
	if (!dev->bulk_in_buffer) 
	{
	  kerror(("altro_driver(altro_probe): Couldn't allocate bulk_in_buffer\n"));
	  goto error;
	}
      }
      else 
	kdebug(("altro_driver(altro_probe): we found yet another bulk in endpoint!\n"));
    }
    else if (!(endpoint->bEndpointAddress & USB_DIR_IN) && ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) == USB_ENDPOINT_XFER_BULK)) 
    {
      /* we found a bulk out endpoint */
      if (!dev->bulk_out_endpointAddr) 
      {
	kdebug(("altro_driver(altro_probe): we found the first bulk out endpoint\n"));
	/* a probe() may sleep and has no restrictions on memory allocations */ 
	dev->write_urb = usb_alloc_urb(0, GFP_KERNEL);
	if (!dev->write_urb) 
	{
	  kerror(("altro_driver(altro_probe): No free urbs available\n"));
	  goto error;
	}
	dev->bulk_out_endpointAddr     = endpoint->bEndpointAddress;
	buffer_size                    = endpoint->wMaxPacketSize;
	dev->bulk_out_size             = buffer_size;
	dev->write_urb->transfer_flags = (URB_NO_TRANSFER_DMA_MAP); 
	dev->bulk_out_buffer           = usb_buffer_alloc(udev, buffer_size, GFP_KERNEL, &dev->write_urb->transfer_dma);
	kdebug(("altro_driver(altro_probe): bulk out endpoint address: %d\n", dev->bulk_out_endpointAddr));
	if (!dev->bulk_out_buffer) 
	{
	  kerror(("altro_driver(altro_probe): Couldn't allocate bulk_out_buffer\n"));
	  goto error;
	}
	usb_fill_bulk_urb(dev->write_urb, udev, usb_sndbulkpipe(udev, endpoint->bEndpointAddress),
			  dev->bulk_out_buffer, buffer_size, altro_write_bulk_callback, dev);
      }
      else 
	kdebug(("altro_driver(altro_probe): we found yet another bulk out endpoint!\n"));
    }
    else 
      kerror(("altro_driver(altro_probe): we found an unidentifiable endpoint!\n"));
  }
  if (!(dev->bulk_in_endpointAddr && dev->bulk_out_endpointAddr)) 
  {
    kerror(("altro_driver(altro_probe): Couldn't find both bulk-in and bulk-out endpoints"));
    goto error;
  }

  dev->present = 1;  /* allow device read, write and ioctl */

  /* we can register the device now, as it is ready */
  usb_set_intfdata(interface, dev);
  retval = usb_register_dev(interface, &altro_class);
  if (retval) 
  {
    /* something prevented us from registering this driver */
    kerror(("altro_driver(altro_probe): Not able to get a minor for this device.\n"));
    usb_set_intfdata (interface, NULL);
    goto error;
  }

  dev->minor = interface->minor;
  add_our_device(dev);
  /* let the user know what node this device is now attached to */
  kdebug(("altro_driver(altro_probe): USB ALICE ALTRO device now attached to USB ALICE ALTRO-%d\n", dev->minor));
  return 0;

 error:
  altro_delete (dev);
  kdebug(("altro_driver(altro_probe): Function done\n"));
  return retval;
}


/***********************************************************/
static void altro_disconnect(struct usb_interface *interface)
/***********************************************************/
{
  struct altro_device *dev;
  int minor;

  kdebug(("altro_driver(altro_disconnect): Function called\n"));
  down(&disconnect_sem);  /* prevent races with open() */
  dev = usb_get_intfdata(interface);
  usb_set_intfdata(interface, NULL);
  down(&dev->sem);
  remove_our_device(dev);
  minor = dev->minor;
  usb_deregister_dev(interface, &altro_class);  /* give back our minor */

  /* terminate an ongoing write */
  if (atomic_read(&dev->write_busy)) 
  {
    usb_unlink_urb(dev->write_urb);
    wait_for_completion(&dev->write_finished);
  }

  dev->present = 0;  /* prevent device read, write and ioctl */
  up(&dev->sem);

  /* if the device is opened, altro_release will clean this up */
  if (!dev->open)
    altro_delete(dev);

  up (&disconnect_sem);

  kdebug(("altro_driver(altro_disconnect): USB ALICE ALTRO #%d now disconnected\n", minor));
  kdebug(("altro_driver(altro_disconnect): Function done\n"));
}


/******************************************************************************************/
static int altro_proc_write(struct file *file, const char *buffer, u_long count, void *data)
/******************************************************************************************/
{
  int len;
  struct altro_proc_data_t *fb_data = (struct altro_proc_data_t *)data;
  
  kdebug(("altro_driver(altro_proc_write): Function called\n"));
  kdebug(("altro_driver(altro_proc_write): user is writing to /proc/altro\n"));

  if (count > 99) 
    len = 99;
  else           
    len = count;

  if (copy_from_user(fb_data->value, buffer, len)) 
  {
    kdebug(("altro_driver(altro_proc_write): error when copying from user\n"));
    return(-EFAULT);
  }

  kdebug(("altro_driver(altro_proc_write): length is %d\n", len));
  fb_data->value[len - 1] = '\0';
  kdebug(("altro_driver(altro_proc_write): text passed is \"%s\"\n", fb_data->value));
 
  if (!strcmp(fb_data->value, "debug")) 
  {
    debug = 1;
    kdebug(("altro_driver(altro_proc_write): debugging enabled\n")); 
  }
 
  if (!strcmp(fb_data->value, "nodebug")) 
  {
    kdebug(("altro_driver(altro_proc_write): debugging disabled\n")); 
    debug = 0; 
  }

  if (!strcmp(fb_data->value, "elog")) 
  {
    errorlog = 1;
    kdebug(("altro_driver(altro_proc_write): error logging enabled\n")); 
  }
 
  if (!strcmp(fb_data->value, "noelog")) 
  {
    kdebug(("altro_driver(altro_proc_write): error logging disabled\n")); 
    errorlog = 0; 
  }
  
  if (!strcmp(fb_data->value, "datadump")) 
  {
    ddump = 1;
    kdebug(("altro_drv(proc_write_altro): data dumping enabled\n")); 
  }
 
  if (!strcmp(fb_data->value, "nodatadump")) 
  {
    ddump = 0; 
    kdebug(("altro_drv(proc_write_altro): data dumping disabled\n")); 
  }
  
  kdebug(("altro_driver(altro_proc_write): Function done\n"));
  return len;
}


/**********************************************************************************************/
static int altro_proc_read(char *page, char **start, off_t off, int count, int *eof, void *data)
/**********************************************************************************************/
{
  int len;
  u_int i;
  
  kdebug(("altro_driver(altro_proc_read): Function called\n"));
  kdebug(("altro_driver(altro_proc_read): user is reading /proc/altro\n"));

  len =  sprintf(page, "ALICE USB-ALTRO driver\n\n");
  len += sprintf(page + len, "Number of devices connected: %d\n", n_our_devices);
  for (i = 0; i < n_our_devices; i++) 
    len += sprintf(page + len, "\tDevice:\t %s (0x%x)\n", our_devices[i]->name, our_devices[i]->minor);

  len += sprintf(page + len, " \n");
  len += sprintf(page + len, "The command 'echo <action> > /proc/altro', executed as root,\n");
  len += sprintf(page + len, "allows you to interact with the driver. Possible actions are:\n");
  len += sprintf(page + len, "debug      -> enable debugging\n");
  len += sprintf(page + len, "nodebug    -> disable debugging\n");
  len += sprintf(page + len, "eolg       -> enable error logging\n");
  len += sprintf(page + len, "noeolg     -> disable error logging\n");    
  len += sprintf(page + len, "datadump   -> enable data dumping (use with care)\n");
  len += sprintf(page + len, "nodatadump -> disable data dumping\n");
 
  *eof = 1;
  *start = page + off;
  len -= off;

  if (len > count) 
    len = count;
  kdebug(("altro_driver(altro_proc_read): Function done\n"));
  return (len);
}


/************************************/
static int __init usb_altro_init(void)
/************************************/
{
  int result;
  
  kdebug(("altro_driver(usb_altro_init): Function called\n"));
  /* Install `/proc' entry */
  altro_proc_file = create_proc_entry("altro", 0644, NULL);
  if (altro_proc_file == NULL) 
  {
    kerror(("altro_driver(usb_altro_init): create_proc_entry failed\n"));
    return -ENOMEM;
  }
  strcpy(altro_proc_data.name, "altro");
  strcpy(altro_proc_data.value, "altro");
  altro_proc_file->data       = &altro_proc_data;
  altro_proc_file->read_proc  = altro_proc_read;
  altro_proc_file->write_proc = altro_proc_write;
  altro_proc_file->owner      = THIS_MODULE;

  /* register this driver with the USB subsystem */
  result = usb_register(&altro_driver);
  if (result) 
  {
    kerror(("altro_driver(usb_altro_init): usb_register failed. Error number = %d\n", result));
    return result;
  }
  
  kdebug(("altro_driver(usb_altro_init): " DRIVER_DESC " " DRIVER_VERSION));
  kdebug(("altro_driver(usb_altro_init): \nALTRO: debug is %d\n", debug));
  kdebug(("altro_driver(usb_altro_init): Function done\n"));
  return 0;
}


/*************************************/
static void __exit usb_altro_exit(void)
/*************************************/
{
  kdebug(("altro_driver(usb_altro_exit): Function called\n"));
  /* deregister this driver with the PROC subsystem */
  remove_proc_entry("altro", NULL);
  
  /* deregister this driver with the USB subsystem */
  usb_deregister(&altro_driver);
  kdebug(("altro_driver(usb_altro_exit): Function done\n"));
}
  
module_init(usb_altro_init);
module_exit(usb_altro_exit);


